/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 06 - solution 1
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

typedef struct cell_s {
  int val;       // Value
  int rowFlag;   // 1 to exclude a value already used
                 // on a row
  int colFlag;   // 1 to exclude a value already used
                 // on a column
  char leftRel;  // Left relationship 
  char upRel;    // Up relationship
} cell_t;

cell_t **futoshiki_read(char *, int *, int *, int *);
int relation_check(cell_t **, int, int);
void futoshiki_write(cell_t **, int);
int futoshiki_solve(cell_t **, int, int);

int main (int argc, char *argv[])
{
  int i, n, valid, complete;
  cell_t **scheme;

  scheme = futoshiki_read(argv[1], &n, &complete, &valid);

  if (complete) {
    fprintf (stdout, "Complete scheme.\n");
    if (valid) {
      fprintf (stdout, "Valid solution.\n");
    } else {
      fprintf (stdout, "Invalid solution.\n");
    }
  } else {
    fprintf (stdout, "Incomplete scheme.\n");
    if (valid) {
      valid = futoshiki_solve(scheme, n, 0);
    } 
    if (valid) {
      fprintf (stdout, "Solution:\n");
      futoshiki_write(scheme, n);
    } else {
      fprintf (stdout, "A solution does not exist.\n");
    }
  }


  for (i=0; i<n; i++) {
    free (scheme[i]);
  }
  free (scheme);

  return EXIT_SUCCESS;
}

/*
 *  Read file and check for correctness
 */
cell_t **futoshiki_read (char *name, int *dim, int *complete, int *valid) {
  cell_t **scheme;
  int n, i, j, val;
  char sep[10];
  FILE *fp;

  fp = fopen (name, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit (EXIT_FAILURE);
  }
  fscanf (fp, "%d", &n);
  scheme = (cell_t **) malloc (n*sizeof(cell_t *));
  if (scheme == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit (EXIT_FAILURE);
  }

  // Matrix allocation and initialization phase
  for (i=0; i<n; i++) {
    scheme[i] = (cell_t *) malloc (n*sizeof(cell_t));
    if (scheme[i] == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit (EXIT_FAILURE);
    }
    for (j=0; j<n; j++) {
      scheme[i][j].val = 0;
      scheme[i][j].leftRel = '|';
      scheme[i][j].upRel = '-';
      scheme[i][j].rowFlag = 0;
      scheme[i][j].colFlag = 0;
    }
  }

  *dim = n;
  *complete = 1;
  *valid = 1;
  for (i=0; i<n; i++) {
    // For all rows, but the first one, read "up" relationship
    if (i > 0) {
      for (j=0; j<n; j++) {
	fscanf (fp, "%s", sep);
	scheme[i][j].upRel = sep[0];
	if (j < n-1) {
	  fscanf (fp, "%*s");
	}
      }
    }

    // For all rows read values and "left" relationship
    for (j=0; j<n; j++) {
      if (j > 0) {
	fscanf (fp, "%s", sep);
	scheme[i][j].leftRel = sep[0];
      }
      fscanf (fp, "%d", &val);
      scheme[i][j].val = val;
      if (val == 0) {
	*complete = 0;
      } else if (val<0 || val>n) {
	*valid = 0;
      } else {
	if (scheme[i][val-1].rowFlag++) {
	  *valid = 0;
	}
	if (scheme[val-1][j].colFlag++) {
	  *valid = 0;
	}
	if (relation_check(scheme, i, j) == 0) {
	  *valid = 0;
	}
      }
    }
  }
  fclose(fp);

  return scheme;
}

/*
 *  Check relationships
 *  return 1 iff OK, 0 otherwise
 */
int relation_check (cell_t **scheme, int row, int col) {
  int val;
  char sep;

  val = scheme[row][col].val;

  // Check condition with the element of the left-hand side (leftRel)
  if (col>0 && scheme[row][col-1].val!=0) {
    sep = scheme[row][col].leftRel;
    if (sep=='<' && scheme[row][col-1].val>val) {
      return 0;
    }
    if (sep=='>' && scheme[row][col-1].val<val) {
      return 0;
    }
  }

  // Check condition with the element of the right-hand side (rightRel)
  if (row>0 && scheme[row-1][col].val!=0) {
    sep = scheme[row][col].upRel;
    if (sep=='^' && scheme[row-1][col].val>val) {
      return 0;
    }
    if (sep=='v' && scheme[row-1][col].val<val) {
      return 0;
    }
  }

  return 1;
}

/*
 *  Solve puzzle: Set all missing values recursively
 */
int futoshiki_solve (cell_t **scheme, int n, int step) {
  int i, j, k;

  // Termination condition
  if (step >= n*n) {
    return 1;
  }

  // From n to row (i) and column (j) indices
  i = step / n; 
  j = step % n;

  // Check current element
  if (scheme[i][j].val) {
    if (relation_check(scheme, i, j)) {
      return futoshiki_solve(scheme, n, step+1);
    } else {
      return 0;
    }
  }

  // Set current element
  for (k=1; k<=n; k++) {
    scheme[i][j].val = k;
    // Value k does not have to be used and must satisfy relationships
    if (scheme[i][k-1].rowFlag==0 && scheme[k-1][j].colFlag==0 &&
      relation_check(scheme, i, j)==1) {
      scheme[i][k-1].rowFlag++;
      scheme[k-1][j].colFlag++;
      if (futoshiki_solve(scheme, n, step+1)) {
	return 1;
      }
      scheme[i][k-1].rowFlag--;
      scheme[k-1][j].colFlag--;
    }
    scheme[i][j].val = 0;
  }
  return 0;
}

/*
 *  Write scheme with the same input format
 */
void futoshiki_write (cell_t **scheme, int n) {
  int i, j;

  for (i=0; i<n; i++) {
    if (i > 0) {
      // For each row, but the first one, print symbols above the row
      for (j=0; j<n; j++) {
	fprintf (stdout, "%c", scheme[i][j].upRel);
	if (j < n-1) {
	  fprintf (stdout, " - ");
	}
      }
      fprintf (stdout, "\n");
    }
    // For each column, but the first one, print symbols on the left
    for (j=0; j<n; j++) {
      if (j > 0) {
	fprintf (stdout, " %c ", scheme[i][j].leftRel);
      }
      fprintf (stdout, "%d", scheme[i][j].val);
    }
    fprintf (stdout, "\n");
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2018
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
